"""
.. module:: test_subprocess_mpi
  :synopsis: Test an external objective function
.. moduleauthor:: David Eriksson <dme65@cornell.edu>
"""

from mpi4py import MPI
import logging
from pySOT import *
from poap.mpiserve import MPIController, MPIProcessWorker
import numpy as np
from subprocess32 import Popen, PIPE
import os.path


def array2str(x):
    return ",".join(np.char.mod('%f', x))


class CppSim(MPIProcessWorker):
    def eval(self, record_id, params, extra_args=None):
        try:
            self.process = Popen(['./sphere_ext', array2str(params[0])], stdout=PIPE)
            val = self.process.communicate()[0]
            self.finish_success(record_id, float(val))
        except ValueError:
            logging.info("WARNING: Incorrect output or crashed evaluation")
            self.finish_cancelled(record_id)


def main_worker():
    logging.basicConfig(filename="./logfiles/test_subprocess_mpi.log",
                        level=logging.INFO)
    CppSim().run()


def main_master(nworkers):
    if not os.path.exists("./logfiles"):
        os.makedirs("logfiles")
    if os.path.exists("./logfiles/test_subprocess_mpi.log"):
        os.remove("./logfiles/test_subprocess_mpi.log")
    logging.basicConfig(filename="./logfiles/test_subprocess_mpi.log",
                        level=logging.INFO)

    print("\nTesting the POAP MPI controller with {0} workers".format(nworkers))
    print("Maximum number of evaluations: 200")
    print("Search strategy: Candidate DYCORS")
    print("Experimental design: Symmetric Latin Hypercube")
    print("Surrogate: Cubic RBF")

    assert os.path.isfile("./sphere_ext"), "You need to build sphere_ext"

    maxeval = 200

    data = Sphere(dim=10)
    print(data.info)

    # Create a strategy and a controller
    strategy = \
        SyncStrategyNoConstraints(
            worker_id=0, data=data,
            maxeval=maxeval, nsamples=nworkers,
            exp_design=SymmetricLatinHypercube(dim=data.dim, npts=2*(data.dim+1)),
            sampling_method=CandidateDYCORS(data=data, numcand=100*data.dim),
            response_surface=RBFInterpolant(kernel=CubicKernel, tail=LinearTail,
                                            maxp=maxeval, dim=data.dim))

    controller = MPIController(strategy)

    # Run the optimization strategy
    result = controller.run()
    print('Best value found: {0}'.format(result.value))
    print('Best solution found: {0}\n'.format(
        np.array_str(result.params[0], max_line_width=np.inf,
                     precision=5, suppress_small=True)))

if __name__ == '__main__':
    # Extract the rank
    comm = MPI.COMM_WORLD
    rank = comm.Get_rank()
    nprocs = comm.Get_size()

    if rank == 0:
        main_master(nprocs)
    else:
        main_worker()



try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO

from experimental_design import *
from rbf import *
from rs_capped import RSCapped, RSUnitbox
from sampling_methods import *
from test_problems import *
from sot_sync_strategies import *
from ensemble_surrogate import EnsembleSurrogate
from poly_regression import *
from merit_functions import *

try:
    from GUI import GUI
    __with_gui__ = True
except ImportError:
    __with_gui__ = False

try:
    from mars_interpolant import MARSInterpolant
    __with_mars__ = True
except ImportError:
    __with_mars__ = False

try:
    from kriging_interpolant import KrigingInterpolant
    __with_kriging__ = True
except ImportError:
    __with_kriging__ = False

__version__ = '0.1.26'
__author__ = 'David Eriksson, David Bindel'



"""
.. module:: test_simple_mpi
  :synopsis: Test Simple MPI
.. moduleauthor:: David Eriksson <dme65@cornell.edu>
"""

from mpi4py import MPI
from pySOT import *
from poap.mpiserve import MPIController, MPISimpleWorker
import numpy as np
import os.path


def main_worker(objfunction):
    MPISimpleWorker(objfunction).run()


def main_master(data, nworkers):
    if not os.path.exists("./logfiles"):
        os.makedirs("logfiles")
    if os.path.exists("./logfiles/test_simple_mpi.log"):
        os.remove("./logfiles/test_simple_mpi.log")
    logging.basicConfig(filename="./logfiles/test_simple_mpi.log",
                        level=logging.INFO)

    print("\nTesting the POAP MPI controller with {0} workers".format(nworkers))
    print("Maximum number of evaluations: 500")
    print("Sampling method: CandidateDYCORS, with weight 0.5")
    print("Experimental design: Symmetric Latin Hypercube")
    print("Surrogate: Cubic RBF, domain scaled to unit box")

    maxeval = 500
    print(data.info)

    # Create a strategy and a controller
    strategy = \
        SyncStrategyNoConstraints(
            worker_id=0, data=data,
            maxeval=maxeval, nsamples=nworkers,
            exp_design=SymmetricLatinHypercube(dim=data.dim, npts=2*(data.dim+1)),
            response_surface=RBFInterpolant(kernel=CubicKernel, tail=LinearTail,
                                            maxp=maxeval, dim=data.dim),
            sampling_method=CandidateDYCORS(data=data, numcand=100*data.dim, weights=[0.5]))
    controller = MPIController(strategy)

    result = controller.run()
    print('Best value found: {0}'.format(result.value))
    print('Best solution found: {0}\n'.format(
        np.array_str(result.params[0], max_line_width=np.inf,
                     precision=5, suppress_small=True)))


if __name__ == '__main__':
    # Optimization problem
    data = Ackley(dim=10)

    # Extract the rank
    comm = MPI.COMM_WORLD
    rank = comm.Get_rank()
    nprocs = comm.Get_size()

    if rank == 0:
        main_master(data, nprocs)
    else:
        main_worker(data.objfunction)



"""
.. module:: test_rbf_regression
  :synopsis: Test RBF regression
.. moduleauthor:: David Eriksson <dme65@cornell.edu>
"""

from pySOT import *
import numpy as np


def main():
    data = Sphere(dim=1)
    maxeval = 100

    kernel = CubicKernel
    tail = LinearTail

    for i in range(2):

        rbf1 = RBFInterpolant(kernel=kernel, tail=tail, maxp=maxeval, dim=data.dim, eta=0)
        rbf2 = RBFInterpolant(kernel=kernel, tail=tail, maxp=maxeval, dim=data.dim, eta='adapt')

        X = np.random.uniform(data.xlow, data.xup, (maxeval, data.dim))
        fX = np.zeros((maxeval, data.dim))
        for j in range(maxeval):
            fX[j] = data.objfunction(X[j, :])
            if i == 0:
                fX[j] += 3*np.cos(10000*X[j, :])
            rbf1.add_point(X[j, :], fX[j])
            rbf2.add_point(X[j, :], fX[j])

        Xpred = np.atleast_2d(np.linspace(data.xlow, data.xup, 1000)).transpose()
        fXpred = np.zeros((1000, data.dim))
        for j in range(1000):
            fXpred[j] = data.objfunction(Xpred[j, :])

        if i == 0:
            print("\nL2 error interpolation with noise: {0}".format(np.linalg.norm(rbf1.evals(Xpred) - fXpred)))
            print("L2 error regularization with noise: {0}".format(np.linalg.norm(rbf2.evals(Xpred) - fXpred)))
        else:
            print("L2 error interpolation without noise: {0}".format(np.linalg.norm(rbf1.evals(Xpred) - fXpred)))
            print("L2 error regularization without noise: {0}\n".format(np.linalg.norm(rbf2.evals(Xpred) - fXpred)))

        """
        import matplotlib.pyplot as plt
        Xpred = np.atleast_2d(np.linspace(data.xlow, data.xup, 1000)).transpose()
        fXpred = np.zeros((1000, data.dim))
        for j in range(1000):
            fXpred[j] = data.objfunction(Xpred[j, :]) + 3*np.cos(10000*Xpred[j, :])

        plt.plot(Xpred, rbf1.evals(Xpred), c='b', linewidth=1.0, zorder=2)
        plt.hold('on')
        plt.plot(Xpred, rbf2.evals(Xpred), c='r', linewidth=1.0, zorder=1)
        plt.scatter(X, fX, c='k', s=10)
        plt.ylim([-35, 45])
        plt.show()
        """

if __name__ == '__main__':
    main()



import numpy as np
import numpy.linalg as la
import scipy.spatial as scpspatial
import scipy.linalg as scplinalg

""" Kernels """


class CubicKernel(object):

    def order(self):
        return 2

    def phiZero(self):
        return 0

    def eval(self, dists):
        return np.multiply(dists, np.multiply(dists, dists))

    def deriv(self, dists):
        return 3 * np.multiply(dists, dists)


class TPSKernel(object):

    def order(self):
        return 2

    def phiZero(self):
        return 0

    def eval(self, dists):
        return np.multiply(np.multiply(dists, dists), np.log(dists + np.finfo(float).tiny))

    def deriv(self, dists):
        return np.multiply(dists, 1 + 2 * np.log(dists + np.finfo(float).tiny))


class LinearKernel(object):

    def order(self):
        return 1

    def phiZero(self):
        return 0

    def eval(self, dists):
        return dists

    def deriv(self, dists):
        return np.ones((dists.shape[0], dists.shape[1]))

""" TAILS """


class LinearTail(object):

    def degree(self):
        return 1

    def dimTail(self, dim):
        return 1 + dim

    # Each row is the basis evaluated at X[i, :]
    def eval(self, X):
        if len(X.shape) == 1:
            X = np.atleast_2d(X)
        return np.hstack((np.ones((X.shape[0], 1)), X))

    # Evaluate the derivative at one point
    # Row i is the gradient wrt x_i
    def deriv(self, x):
        return np.hstack((np.zeros((len(x), 1)), np.eye((len(x)))))


class ConstantTail(object):

    def degree(self):
        return 0

    def dimTail(self, dim):
        return 1

    # Each row is the basis evaluated at X[i, :]
    def eval(self, X):
        if len(X.shape) == 1:
            X = np.atleast_2d(X)
        return np.ones((X.shape[0], 1))

    # Evaluate the derivative at one point
    # Row i is the gradient wrt x_i
    def deriv(self, x):
        return np.ones((len(x), 1))


""" RBF Interpolant """


class RBFInterpolant(object):

    def __init__(self, kernel=None, tail=None, maxp=500, dim=None, eta=1e-8):
        if kernel is None or tail is None:
            kernel = CubicKernel
            tail = LinearTail

        self.maxp = maxp
        self.nump = 0
        self.kernel = kernel()
        self.tail = tail()
        self.ntail = self.tail.dimTail(dim)
        self.A = None
        self.LU = None
        self.piv = None
        self.c = None
        self.dim = dim
        self.x = None
        self.fx = None
        self.rhs = None
        self.c = None
        self.eta = eta
        self.updated = False

        if eta is not 'adapt' and (eta < 0 or eta >= 1):
            raise ValueError("eta has to be in [0,1) or be the string 'adapt' ")

        if self.kernel.order() - 1 > self.tail.degree():
            raise ValueError("Kernel and tail mismatch")

    def reset(self):
        """Re-set the interpolation."""
        self.nump = 0
        self.x = None
        self.fx = None
        self.rhs = None
        self.A = None
        self.LU = None
        self.piv = None
        self.c = None
        self.updated = False

    def _alloc(self, dim, ntail):
        """Allocate storage for x, fx, rhs, and A.

        :param dim: Number of dimensions
        :param ntail: Number of tail functions
        """
        maxp = self.maxp
        self.dim = dim
        self.ntail = ntail
        self.x = np.zeros((maxp, dim))
        self.fx = np.zeros((maxp, 1))
        self.rhs = np.zeros((maxp+ntail, 1))
        self.A = np.zeros((maxp+ntail, maxp+ntail))

    def _realloc(self, dim, ntail, extra=1):
        """Expand allocation to accommodate more points (if needed)

        :param dim: Number of dimensions
        :param ntail: Number of tail functions
        :param extra: Number of additional points to accommodate
        """
        if self.nump == 0:
            self._alloc(dim, ntail)
        elif self.nump + extra > self.maxp:
            self.maxp = max(self.maxp*2, self.maxp+extra)
            self.x.resize((self.maxp, dim))
            self.fx.resize((self.maxp, 1))
            self.rhs.resize((self.maxp+ntail, 1))
            A0 = self.A  # pylint: disable=invalid-name
            self.A = np.zeros((self.maxp+ntail, self.maxp+ntail))
            self.A[:A0.shape[0], :A0.shape[1]] = A0

    def coeffs(self):
        """Compute the expansion coefficients

        :return: Expansion coefficients
        """

        if self.c is None:
            nact = self.ntail + self.nump

            if self.eta is 'adapt':
                eta_vec = np.linspace(0, 0.99, 30)
            else:
                eta_vec = np.array([self.eta])

            rms_best = np.inf

            for i in range(len(eta_vec)):
                eta = eta_vec[i]

                Aa = np.copy(self.A[:nact, :nact])
                for j in range(self.nump):
                    Aa[j + self.ntail, j + self.ntail] += eta/(1-eta)*self.nump

                [LU, piv] = scplinalg.lu_factor(Aa)
                c = scplinalg.lu_solve((LU, piv), self.rhs[:nact])

                # Do LOOCV if requested
                if self.eta is 'adapt':
                    I = np.eye(nact)
                    AinvI = scplinalg.lu_solve((LU, piv), I[:, self.ntail:])

                    chat = c - np.multiply(AinvI, np.transpose(
                        c[self.ntail:]/np.transpose(np.atleast_2d(np.diag(AinvI[self.ntail:, :])))))

                    for j in range(self.nump):
                        chat[j + self.ntail, j] = 0

                    f_pred = np.sum(np.transpose(self.A[self.ntail:nact, self.ntail:nact]) * chat[self.ntail:, :], axis=0) + \
                        np.sum(np.transpose(self.A[self.ntail:nact, :self.ntail]) * chat[:self.ntail, :], axis=0)
                    rms_val = np.sqrt(np.sum((self.fx[:self.nump] - np.transpose(np.atleast_2d(f_pred))) ** 2)/self.nump)

                    if rms_val < rms_best:
                        rms_best = rms_val
                        self.eta_best = eta
                        self.c = np.copy(c)
                        self.piv = piv
                        self.LU = LU
                else:
                    self.c = c
                    self.piv = piv
                    self.LU = LU
                    return self.c

        return self.c

    def get_x(self):
        """Get the list of data points

        :return: List of data points
        """
        return self.x[:self.nump, :]

    def get_fx(self):
        """Get the list of function values for the data points.

        :return: List of function values
        """
        return self.fx[:self.nump, :]

    def add_point(self, xx, fx):
        """Add a new function evaluation

        :param xx: Point to add
        :param fx: The function value of the point to add
        """

        self._realloc(self.dim, self.ntail)

        self.x[self.nump, :] = xx
        self.fx[self.nump] = fx
        self.rhs[self.ntail + self.nump] = fx

        self.nump += 1
        nact = self.nump + self.ntail

        p = self.tail.eval(xx)
        phi = self.kernel.eval(scpspatial.distance.cdist(self.get_x(), np.atleast_2d(xx)))

        #  Create the matrix with the initial points
        self.A[nact-1, 0:self.ntail] = p.ravel()
        self.A[0:self.ntail, nact-1] = p.ravel()
        self.A[nact-1, self.ntail:nact] = phi.ravel()
        self.A[self.ntail:nact, nact-1] = phi.ravel()

        # Coefficients and LU are outdated
        self.LU = None
        self.piv = None
        self.c = None

        self.updated = False

    def transform_fx(self, fx, d=None):
        """Replace f with transformed function values for the fitting

        :param fx: Transformed function value
        """
        self.rhs[self.ntail:self.ntail+self.nump] = fx
        self.LU = None
        self.piv = None
        self.c = None

    def eval(self, x, ds=None):
        """Evaluate the rbf interpolant at the point x

        :param xx: Point where to evaluate
        :return: Value of the rbf interpolant at x
        """

        px = self.tail.eval(x)
        ntail = self.ntail
        c = self.coeffs()
        if ds is None:
            ds = scpspatial.distance.cdist(np.atleast_2d(x), self.x[:self.nump, :])
        fx = np.dot(px, c[:ntail]) + np.dot(self.kernel.eval(ds), c[ntail:ntail+self.nump])
        return fx[0][0]

    def evals(self, x, ds=None):
        """Evaluate the rbf interpolant at the points x

        :param xx: Points where to evaluate
        :return: Values of the rbf interpolant at x
        """
        ntail = self.ntail
        c = np.asmatrix(self.coeffs())
        if ds is None:
            ds = scpspatial.distance.cdist(x, self.x[:self.nump, :])
        fx = self.kernel.eval(ds)*c[ntail:ntail+self.nump] + self.tail.eval(x)*c[:ntail]
        return fx

    def deriv(self, x):
        """Evaluate the derivative of the rbf interpolant at x

        :param x: Data point
        :return: Derivative of the rbf interpolant at x (of size 1 x dim)
        """

        if len(x.shape) == 1:
            x = np.atleast_2d(x)  # Make x 1-by-dim
        ntail = self.ntail
        dpx = self.tail.deriv(x.transpose())
        c = self.coeffs()
        dfx = np.dot(dpx, c[:ntail]).transpose()
        ds = scpspatial.distance.cdist(self.x[:self.nump, :], x)
        ds[ds < 1e-10] = 1e-10  # Better safe than sorry
        dsx = - self.x[:self.nump, :]
        dsx += x
        dsx *= (np.multiply(self.kernel.deriv(ds), c[ntail:]) / ds)
        dfx += np.sum(dsx, 0)

        return dfx



"""
.. module:: test_subprocess_partial_info
  :synopsis: Test an external objective function with partial info
.. moduleauthor:: David Eriksson <dme65@cornell.edu>
"""

import logging
from pySOT import *
from poap.controller import ThreadController, ProcessWorkerThread
import numpy as np
from subprocess32 import Popen, PIPE
import os.path


def array2str(x):
    return ",".join(np.char.mod('%f', x))


class SumfunExt:
    def __init__(self, dim=10):
        self.xlow = -5 * np.ones(dim)
        self.xup = 5 * np.ones(dim)
        self.dim = dim
        self.info = str(dim)+"-dimensional Sumfun function \n" +\
                             "Global optimum: f(0,0,...,0) = 0"
        self.min = 0
        self.integer = []
        self.continuous = np.arange(0, dim)


class CppSim(ProcessWorkerThread):

    def handle_eval(self, record):
        self.process = Popen(['./sumfun_ext', array2str(record.params[0])],
                             stdout=PIPE)

        val = np.nan
        # Continuously check for new outputs from the subprocess
        while True:
            output = self.process.stdout.readline()
            if output == '' and self.process.poll() is not None:  # No new output
                break
            if output:  # New intermediate output
                try:
                    val = float(output.strip())  # Try to parse output
                    if val > 350:  # Terminate if too large
                        self.process.terminate()
                        self.finish_success(record, 350)
                        return
                except ValueError:  # If the output is nonsense we terminate
                    logging.warning("Incorrect output")
                    self.process.terminate()
                    self.finish_cancelled(record)
                    return

        rc = self.process.poll()  # Check the return code
        if rc < 0 or np.isnan(val):
            logging.info("WARNING: Incorrect output or crashed evaluation")
            self.finish_cancelled(record)
        else:
            self.finish_success(record, val)


def main():
    if not os.path.exists("./logfiles"):
        os.makedirs("logfiles")
    if os.path.exists("./logfiles/test_subprocess_partial_info.log"):
        os.remove("./logfiles/test_subprocess_partial_info.log")
    logging.basicConfig(filename="./logfiles/test_subprocess_partial_info.log",
                        level=logging.INFO)

    print("\nNumber of threads: 4")
    print("Maximum number of evaluations: 200")
    print("Sampling method: Candidate DYCORS")
    print("Experimental design: Symmetric Latin Hypercube")
    print("Surrogate: Cubic RBF")

    assert os.path.isfile("./sumfun_ext"), "You need to build sumfun_ext"

    nthreads = 4
    maxeval = 200
    nsamples = nthreads

    data = SumfunExt(dim=10)
    print(data.info)

    # Create a strategy and a controller
    controller = ThreadController()
    controller.strategy = \
        SyncStrategyNoConstraints(
            worker_id=0, data=data,
            maxeval=maxeval, nsamples=nsamples,
            exp_design=SymmetricLatinHypercube(dim=data.dim, npts=2*(data.dim+1)),
            sampling_method=CandidateDYCORS(data=data, numcand=100*data.dim),
            response_surface=RBFInterpolant(kernel=CubicKernel, tail=LinearTail,
                                            maxp=maxeval, dim=data.dim))

    # Launch the threads and give them access to the objective function
    for _ in range(nthreads):
        controller.launch_worker(CppSim(controller))

    # Run the optimization strategy
    result = controller.run()

    print('Best value found: {0}'.format(result.value))
    print('Best solution found: {0}\n'.format(
        np.array_str(result.params[0], max_line_width=np.inf,
                     precision=5, suppress_small=True)))

if __name__ == '__main__':
    main()
